import logging
import time
from timeit import default_timer as timer
import cv2
import asyncio
import sys

from gspeerconnection.gspeerconnectionwatcher import FrameBufferThread


class GSHLSPlayer:

    def __init__(self, gsdbs, onframe, ondestruct, url):
        self.gsdbs = gsdbs
        self.target = "hls_playlist"
        self.onframe = onframe
        self.ondestruct = ondestruct
        self.url = url
        self.url += f"&vision=true&session={self.gsdbs.cookiejar.get('session')}&signature={self.gsdbs.cookiejar.get('signature')}"
        self._logger = logging.getLogger(__name__)
        self.logger = logging.getLogger(__name__)
        self.broadcastthread = FrameBufferThread(self.gsdbs, self.target, self.onframe)
        self.broadcastthread.start()

    async def startrecording(self):
        cap = cv2.VideoCapture(self.url)
        if cap.isOpened() == False:
            self._logger.error("unable to open playlist")
            return
        fps = cap.get(cv2.CAP_PROP_FPS)
        # fps = 1000
        wait_ms = (1000 / fps)
        self._logger.info(f"playing back HLS-Playlist with {fps} fps")
        framecount = 0
        while True:
            ret, frame = cap.read()
            if not ret:
                break

            start_time = time.time()

            self.broadcastthread.write(frame)

            end_time = time.time()

            execution_time_ms = (end_time - start_time) * 1000

            calculatedwait = wait_ms - execution_time_ms / 1000
            # self.logger.info(f"frame written with id {framecount}")
            framecount = framecount + 1
            if calculatedwait > 0:
                time.sleep(calculatedwait / 1000)

        self.ondestruct(self.gsdbs)

        cap.release()


class GSHLSPlayerNew:

    def __init__(self, gsdbs, onframe, ondestruct, url, queue):
        self.gsdbs = gsdbs
        self.target = "hls_playlist"
        self.onframe = onframe
        self.ondestruct = ondestruct
        self.url = url
        self.url += f"&vision=true&session={self.gsdbs.cookiejar.get('session')}&signature={self.gsdbs.cookiejar.get('signature')}"
        self._logger = logging.getLogger(__name__)
        self.logger = logging.getLogger(__name__)
        self.queue = queue
        # self.broadcastthread = FrameBufferThread(self.gsdbs, self.target, self.onframe)
        # self.broadcastthread.start()

    async def startrecording(self):
        cap = cv2.VideoCapture(self.url)
        if cap.isOpened() == False:
            self._logger.error("unable to open playlist")
            return
        fps = cap.get(cv2.CAP_PROP_FPS)
        # fps = 1000
        wait_ms = (1000 / fps)
        self._logger.info(f"playing back HLS-Playlist with {fps} fps")
        while True:
            ret, frame = cap.read()
            if not ret:
                break

            start_time = time.time()

            await self.queue.put(frame)

            end_time = time.time()
            execution_time_ms = (end_time - start_time) * 1000
            calculatedwait = wait_ms - execution_time_ms / 1000

            if calculatedwait > 0:
                await asyncio.sleep(calculatedwait / 1000)

        self.ondestruct(self.gsdbs)

        cap.release()


async def checkHLSQueue(opqueue, gsdbs, source, onframe):
    while True:
        data = await opqueue.get()
        if data is not None:
            await onframe(gsdbs, source, data, 0)


async def createHLSPlayer(gsdbs, onframe, source, ondestruct, playlisturl):
    hlsqueue = asyncio.Queue(3)
    hlsplayer = GSHLSPlayerNew(gsdbs, onframe=onframe, ondestruct=ondestruct, url=playlisturl, queue=hlsqueue)
    hlstask = await asyncio.gather(hlsplayer.startrecording(),
                                   checkHLSQueue(hlsqueue, gsdbs, source, onframe))
    return hlstask



import asyncio
from datetime import datetime, timedelta
from fractions import Fraction
from io import BytesIO

import aiohttp
import av


class VideoRecorderFrameBased:
    def __init__(self, gsdbs, target, rate, eventloop):
        self.segment = None
        self.writer = None
        self.gsdbs = gsdbs
        self.target = target
        self.rate = rate
        self.framecounter = 0
        self.filename = None
        self.duration = 0
        self.eventloop = eventloop
        self.is_recording = False

    def start_recording(self, filename):
        self.segment = av.open(BytesIO(), mode='wb', format='mpegts')
        self.writer = self.segment.add_stream('libx264', rate=self.rate)
        self.filename = filename
        self.writer.time_base = Fraction(1, self.rate)
        self.is_recording = True

    def write(self, frame):
        frame.time_base = Fraction(1, self.rate)
        frame.pts = self.framecounter
        self.duration = frame.time
        self.framecounter = self.framecounter + 1
        self.segment.mux(self.writer.encode(frame))

    def stop_recording(self):
        self.is_recording = False
        self.segment.mux(self.writer.encode())
        self.segment.close()
        self.writer = None
        self.framecounter = 0
        segment_data = self.segment.file
        voddate = datetime.utcnow() - timedelta(seconds=self.duration)
        asyncio.run_coroutine_threadsafe(self.saveVodSub(self.duration,
                                                         self.gsdbs,
                                                         segment_data.getbuffer(),
                                                         voddate,
                                                         self.target,
                                                         f"{self.filename}.ts"),
                                         self.eventloop)

    async def saveVodSub(self, duration, gsdbs, gsvoddata, datetime, target, vodFileName):
        storageURL = gsdbs.credentials["baseurl"] + gsdbs.credentials['storageurl'] + f"/vod/{target}"
        data = aiohttp.FormData()
        data.add_field('filePart', gsvoddata, filename=f"{vodFileName}")
        async with aiohttp.ClientSession() as session:
            async with session.post(storageURL, cookies=gsdbs.cookiejar, data=data) as response:
                response.raise_for_status()
        vodstmtn = f"""
                        mutation{{
                        addDTable(
                            dtablename:"gsvod",
                            superDTable:[DTABLE],
                            sriBuildInfo:"${{streamkey}}-${{fragmentid}}",
                            dataLinks:[{{ alias:"streamkey",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"fragmentid",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"starttime",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:DATETIME}},
                            {{ alias:"segmentlength",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}],
                            data:[["streamkey","fragmentid","starttime","segmentlength"],
                                  ["{target}","{vodFileName}","{datetime}","{duration}"]
                                  ])
                    }}
                    """
        await gsdbs.asyncExecuteStatement(vodstmtn)



import ast
import asyncio
import json
import logging
import subprocess
from threading import Thread
import socketio
from aiortc import RTCPeerConnection, RTCRtpSender, RTCConfiguration, RTCIceServer, RTCSessionDescription
from aiortc.contrib.media import MediaPlayer, MediaRelay
from datetime import datetime

try:
    import picamera
    from picamera.array import PiRGBArray
except:
    log = logging.getLogger(__name__)
    log.error("failed to import picamera")

from gspeerconnection.pitrack import PiH264Relay, PiH264CameraOutput

video_trackList = list()
indexmap = {}
pcs = set()
relay = None
microfon = None




class GSPeerConnectionBroadcasterPiCaller:
    def __init__(self, gsdbs, camera, oncommand=None, setChannel=None):
        self.sio = socketio.AsyncClient()
        self.gsdbs = gsdbs
        self.camera = camera
        self._logger = logging.getLogger(__name__)
        self.oncommand = oncommand
        self.setChannel = setChannel

    async def startSocket(self):
        self._logger.debug("start broadcast socket")
        await GSPeerConnectionBroadcasterPiSocketv3.create(self.sio, self.gsdbs, self.camera, self.oncommand,
                                                           self.setChannel)


def channel_log(logger, channel, t, message):
    logger.info("channel(%s) %s %s" % (channel.label, t, message))


def channel_send(logger, channel, message):
    channel_log(logger, channel, ">", message)
    channel.send(message)


class GSPeerConnectionBroadcasterPiSocketv3:

    def getPiH264Relay(self, id):
        video_track = PiH264Relay(self.gsdbs.credentials["framerate"])

        video_trackList.append(video_track)
        indexmap[id] = video_trackList.index(video_track)
        return video_track

    def getMicrofonRelay(self):
        global relay, microfon
        if relay is None:
            microfon = MediaPlayer("hw:1", format="alsa", options={
                "channels": "1",
                "sample_rate": "44100"
            })
            relay = MediaRelay()
        return relay.subscribe(microfon.audio)

    async def capture_frame(self, frame_queue):
        # Create a buffer for the frame data
        raw_capture = PiRGBArray(self.camera, size=self.camera.resolution)

        # Capture a frame from the camera
        self.camera.capture(raw_capture, format="h264", use_video_port=True)
        frame_data = bytes(raw_capture.array)

        # Put the frame data into the queue
        await frame_queue.put(frame_data)

    async def capture_frames(self):

        self.camera.resolution = (self.gsdbs.credentials["hres"], self.gsdbs.credentials["vres"])
        self.camera.framerate = self.gsdbs.credentials["framerate"]
        asyncio.sleep(1)

        frame_queue = asyncio.Queue(3)

        while True:
            await self.capture_frame(frame_queue)

    async def process_frames(self, frame_queue):
        while True:
            frame_data = await frame_queue.get()

            for track in video_trackList:
                track.write(frame_data)

            frame_queue.task_done()

    @classmethod
    async def create(cls, sio, gsdbs, camera, oncommand, setChannel=None):
        self = GSPeerConnectionBroadcasterPiSocketv3()

        self.rtconfiList = []
        self._logger = logging.getLogger(__name__)
        self.gsdbs = gsdbs
        self.oncommand = oncommand
        self.camera = camera
        self.setChannel = setChannel
        # picam.append(camera)
        if self.gsdbs.credentials["stunenable"]:
            self.rtconfiList.append(RTCIceServer(self.gsdbs.credentials["stunserver"]))
        if self.gsdbs.credentials["turnenable"]:
            for turnserver in self.gsdbs.credentials["turnserver"]:
                self.rtconfiList.append(RTCIceServer(urls=turnserver,
                                                     username=self.gsdbs.credentials["turnuser"],
                                                     credential=self.gsdbs.credentials["turnpw"]
                                                     ))

        self.sio = sio
        self.peerConnections = {}

        self.webcam = None
        self.relay = None

        @self.sio.event
        async def connect():
            self._logger.info('connection established')

        @self.sio.event
        async def wjoined(id):
            await self.sio.emit("joined", self.sio.sid)

        @self.sio.event
        async def broadcaster(id, description, emitter):

            if len(self.rtconfiList) > 0:
                pc = RTCPeerConnection(configuration=RTCConfiguration(self.rtconfiList))
            else:
                pc = RTCPeerConnection()

            self.peerConnections[id] = pc

            video = self.getPiH264Relay(id)
            audio = None

            transceiver = pc.addTransceiver("video", direction="sendonly")

            capabilities = RTCRtpSender.getCapabilities("video")
            preferences = list(filter(lambda x: x.name == "H264", capabilities.codecs))
            preferences += list(filter(lambda x: x.name == "rtx", capabilities.codecs))
            transceiver.setCodecPreferences(preferences)

            if self.gsdbs.credentials["audio"]:
                pc.addTransceiver("audio", direction="sendonly")
                audio = self.getMicrofonRelay()
                pc.addTrack(audio)

            for t in pc.getTransceivers():
                if t.kind == "video":
                    pc.addTrack(video)

            sendchannel = pc.createDataChannel("sendChannel")

            if setChannel is not None:
                setChannel(sendchannel)
            channel_log(self._logger, sendchannel, "-", "created by local party")

            @sendchannel.on("open")
            def on_open():
                self._logger.info("datachannel open")

            @sendchannel.on("message")
            def on_message(message):

                def subprocess_cmd(command):
                    process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
                    proc_stdout = process.communicate()[0].strip()
                    channel_send(self._logger, sendchannel, json.dumps({"response": proc_stdout.decode()}))

                channel_log(self._logger, sendchannel, "<", message)
                if isinstance(message, str) and message.startswith("ping"):
                    channel_send(self._logger, sendchannel, f'{{"ping":"pong","ts":"{datetime.utcnow()} UTC"}}')
                if isinstance(message, str) and message.startswith('{"type'):
                    self.oncommand(self.gsdbs, json.loads(message))
                if isinstance(message, str) and message.startswith('{"command'):
                    subprocess_cmd(json.loads(message)["command"])

            await pc.setLocalDescription(await pc.createOffer())

            await self.sio.emit("watcher", {"target": description,
                                            "id": id,
                                            "sdp": json.dumps(
                                                {"type": pc.localDescription.type,
                                                 "sdp": pc.localDescription.sdp})})

        @self.sio.event
        async def answer(id, description):

            if id in self.peerConnections:
                if self.peerConnections[id].signalingState != "stable":
                    if type(description) == str:
                        description = ast.literal_eval(description)
                    desc = type('new_dict', (object,), description)
                    answer = RTCSessionDescription(sdp=description["sdp"], type=description["type"])
                    await self.peerConnections[id].setRemoteDescription(answer)

        @self.sio.event
        async def disconnectBroadcaster(id, desc):
            if id in self.peerConnections:
                self._logger.info("Disconnected Peer")
                await self.peerConnections[id].close()
                self.peerConnections.pop(id, None)

        @self.sio.event
        async def disconnect():
            self._logger.info('disconnected from server')

        connectURL = ""
        if "localhost" in self.gsdbs.credentials["signalserver"]:
            connectURL = f'{self.gsdbs.credentials["signalserver"]}:{str(self.gsdbs.credentials["signalport"])}'
        else:
            connectURL = self.gsdbs.credentials["signalserver"]

        frame_queue = asyncio.Queue()
        capture_task = asyncio.create_task(self.capture_frames())
        process_task = asyncio.create_task(self.process_frames(frame_queue))

        await self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}{self.gsdbs.credentials["cnode"]}')

        await asyncio.gather(await self.sio.wait(), capture_task, process_task)



import ast
import asyncio
import threading
import json
import logging
from strenum import StrEnum
import socketio
from aiortc import MediaStreamTrack, RTCPeerConnection, RTCIceServer, RTCConfiguration, RTCSessionDescription
from aiortc.contrib.media import MediaRelay, MediaBlackhole
from aiortc.mediastreams import MediaStreamError
from aiortc.rtcrtpreceiver import RemoteStreamTrack
from queue import Queue
from gspeerconnection.gsmediarecorder import GSMediaRecorder
from threading import Thread
import time

relay = MediaRelay()


def synchronize_async_helper(to_await):
    async_response = []

    async def run_and_capture_result():
        r = await to_await
        async_response.append(r)

    try:
        loop = asyncio.get_event_loop()
    except:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    coroutine = run_and_capture_result()
    loop.run_until_complete(coroutine)
    return async_response[0]


class GSCOMMAND(StrEnum):
    PAN = 'pan',
    TILT = 'tilt',
    ZOOM = 'zoom',
    FOCUS = 'focus',
    SNAPSHOT = 'snapshot',
    IRCUT = 'ircut'


class GSPeerConnectionWatcherNew:

    def __init__(self):
        self.sio = None
        self.oncommand = None

    async def sendCommand(self, command: GSCOMMAND, value):
        await self.sio.emit("command", {"type": command, "value": value})

    async def process_frames(self, frame_queue, framerate):
        self.logger.warning(f"starting process frames")
        start_time = time.time()
        frame_count = 0
        fps = framerate
        while True:
            frame_count += 1
            elapsed_time = time.time() - start_time
            if elapsed_time > 1 / fps:
                current_fps = frame_count / elapsed_time
                # print(f'Async Loop FPS: {elapsed_time:.3f}')
                start_time = time.time()
                frame_count = 0
                frame_data = await frame_queue.get()
                self.onframe(self.gsdbs, self.target, frame_data, 0)
                frame_queue.task_done()
            await asyncio.sleep(0)

    @classmethod
    async def create(cls, connectionWatcher, gsdbs, target, onframe=None,  on_data_channel=None,
                     on_message=None, onmessage=None, ontrack=None, sync=True,
                     framerate=15, ontrackended=None, signalingserver=None):
        self = connectionWatcher
        self.rtconfiList = []
        self.sio = socketio.AsyncClient()
        self.peerConnections = None
        self.gsmediaconsumer = None
        self.gsdbs = gsdbs
        self.onframe = onframe
        self.target = target
        self.on_data_channel = on_data_channel
        self.on_message = on_message
        self.ontrack = ontrack
        self.ontrackended = ontrackended
        self.signalingserver = signalingserver
        self.sync = sync
        self.framerate = framerate
        self.logger = logging.getLogger(__name__)

        if self.gsdbs.credentials["stunenable"]:
            self.rtconfiList.append(RTCIceServer(self.gsdbs.credentials["stunserver"]))
        if self.gsdbs.credentials["turnenable"]:
            self.rtconfiList.append(RTCIceServer(self.gsdbs.credentials["turnserver"],
                                                 self.gsdbs.credentials["turnuser"],
                                                 self.gsdbs.credentials["turnpw"]))

        @self.sio.event
        async def connect():
            self.logger.info('connection established')

        @self.sio.event
        async def bjoined(id, type):
            self.logger.info("broadcaster received")
            await self.sio.emit("broadcaster", {"id": id, "target": self.target, "emitter": "python"})

        @self.sio.event
        async def joined(id):
            self.logger.info("calling broadcaster")
            await self.sio.emit("broadcaster", {"id": id, "target": self.target, "emitter": "watcher"})

        @self.sio.event
        async def broadcaster(id, target, emitter):
            if emitter == "watcher" or emitter == "js":
                return

            if len(self.rtconfiList) > 0:
                self.peerConnections = RTCPeerConnection(configuration=RTCConfiguration(self.rtconfiList))
            else:
                self.peerConnections = RTCPeerConnection()

            self.peerConnections.addTransceiver('video', direction='recvonly')

            @self.peerConnections.on("iceconnectionstatechange")
            async def on_iceconnectionstatechange():
                if self.peerConnections.iceConnectionState == "complete":
                    pass

                if self.peerConnections.iceConnectionState == "failed":
                    await self.peerConnections.close()

            @self.peerConnections.on("track")
            async def on_track(track):
                if track.kind == "video":
                    if self.ontrack is not None:
                        self.logger.info("on track received.Recording started")
                        await self.ontrack(self.gsdbs, track, self.target)
                    if self.onframe is not None:
                        self.logger.info("on track received. onframe started")
                        if self.sync:
                            gsmediaconsumer = GSMediaConsumer(self.gsdbs, self.target, self.onframe, self.framerate)
                        else:
                            gsmediaconsumer = GSMediaConsumerAsync(self.gsdbs, self.target, self.onframe,
                                                                   self.frame_queue)
                        gsmediaconsumer.addTrack(track)
                        await gsmediaconsumer.start()

            @self.peerConnections.on("track")
            async def on_track(track):
                if track.kind == "video":
                    if self.ontrack is not None:
                        self.logger.info("on track received.Recording started")
                        await self.ontrack(self.gsdbs, track, self.target)
                    if self.onframe is not None:
                        self.logger.info("on track received. onframe started")
                        if self.sync:
                            gsmediaconsumer = GSMediaConsumer(self.gsdbs, self.target, self.onframe, self.framerate)
                        else:
                            gsmediaconsumer = GSMediaConsumerAsync(self.gsdbs, self.target, self.onframe,
                                                                   self.frame_queue)
                        gsmediaconsumer.addTrack(track)
                        await gsmediaconsumer.start()

            @self.peerConnections.on("datachannel")
            def on_datachannel(channel):
                self.on_data_channel(channel)

                @channel.on("message")
                def on_message(message):
                    self.on_message(message)


            await self.peerConnections.setLocalDescription(await self.peerConnections.createOffer())
            await self.sio.emit("watcher",
                                {"target": self.target,
                                 "id": id,
                                 "sdp": {"type": self.peerConnections.localDescription.type,
                                         "sdp": self.peerConnections.localDescription.sdp}})

        @self.sio.event
        async def watcher(id, description):
            if self.peerConnections is None:
                if len(self.rtconfiList) > 0:
                    self.peerConnections = RTCPeerConnection(configuration=RTCConfiguration(self.rtconfiList))
                else:
                    self.peerConnections = RTCPeerConnection()
                if type(description) == str:
                    description = ast.literal_eval(description)

                offer = RTCSessionDescription(sdp=description["sdp"], type=description["type"])

                self.peerConnections.addTransceiver('video', direction='recvonly')

                @self.peerConnections.on("track")
                async def on_track(track):
                    if track.kind == "video":
                        if self.ontrack is not None:
                            self.logger.info("on track received.Recording started")
                            await self.ontrack(self.gsdbs, track, self.target)
                        if self.onframe is not None:
                            self.logger.info("on track received. onframe started")
                            if self.sync:
                                gsmediaconsumer = GSMediaConsumer(self.gsdbs, self.target, self.onframe, self.framerate)
                            else:
                                gsmediaconsumer = GSMediaConsumerAsync(self.gsdbs, self.target, self.onframe,
                                                                       self.frame_queue)
                            gsmediaconsumer.addTrack(track)
                            await gsmediaconsumer.start()

                @self.peerConnections.on("track")
                async def on_track(track):
                    if track.kind == "video":
                        if self.ontrack is not None:
                            self.logger.info("on track received.Recording started")
                            await self.ontrack(self.gsdbs, track, self.target)
                        if self.onframe is not None:
                            self.logger.info("on track received. onframe started")
                            if self.sync:
                                gsmediaconsumer = GSMediaConsumer(self.gsdbs, self.target, self.onframe, self.framerate)
                            else:
                                gsmediaconsumer = GSMediaConsumerAsync(self.gsdbs, self.target, self.onframe,
                                                                       self.frame_queue)
                            gsmediaconsumer.addTrack(track)
                            await gsmediaconsumer.start()

                @self.peerConnections.on("datachannel")
                def on_datachannel(channel):
                    self.on_data_channel(channel)

                    @channel.on("message")
                    def on_message(message):
                        self.on_message(message)


                await self.peerConnections.setRemoteDescription(offer)
                await self.peerConnections.setLocalDescription(await self.peerConnections.createAnswer())
                await self.sio.emit("answer",
                                    {"id": id,
                                     "message": {"type": self.peerConnections.localDescription.type,
                                                 "sdp": self.peerConnections.localDescription.sdp}})

        @self.sio.event
        async def answer(id, description):
            if self.peerConnections.signalingState != "stable":
                if isinstance(description, dict):
                    desc = type('new_dict', (object,), description)
                else:
                    desc = type('new_dict', (object,), ast.literal_eval(description))
                await self.peerConnections.setRemoteDescription(desc)

        @self.sio.event
        async def disconnect():
            self.logger.info("Socket connection list")
            if self.peerConnections != None:
                await self.peerConnections.close()
            if self.gsmediaconsumer != None:
                await self.gsmediaconsumer.stop()
            await self.ontrackended(self.gsdbs, target)

        @self.sio.event
        async def disconnectPeer(id):
            self.logger.info("Camera connection list")
            if self.peerConnections != None:
                await self.peerConnections.close()
            if self.gsmediaconsumer != None:
                await self.gsmediaconsumer.stop()
            await self.ontrackended(self.gsdbs, target)

        if self.signalingserver is not None:
            connectURL = self.signalingserver

        else:
            if "localhost" in self.gsdbs.credentials["signalserver"]:
                connectURL = f'{self.gsdbs.credentials["signalserver"]}:{str(self.gsdbs.credentials["signalport"])}'
            else:
                connectURL = self.gsdbs.credentials["signalserver"]

        await self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}{self.gsdbs.credentials["cnode"]}&target={self.target}&terminator=true')

        if self.sync:
            await self.sio.wait()
        else:
            self.frame_queue = asyncio.Queue(3)
            process_task = asyncio.create_task(self.process_frames(self.frame_queue, self.framerate))
            return await asyncio.gather(self.sio.wait(), process_task)


class FrameBufferThread(Thread):
    def __init__(self, gsdbs, target, onframe, framerate):
        super(FrameBufferThread, self).__init__()
        self.gsdbs = gsdbs
        self.logger = logging.getLogger(__name__)
        self.logger.debug("Starting Buffer")
        self.target = target
        self.onframe = onframe
        self.framerate = framerate
        self.nal_queue = Queue(3)

    def write(self, frame):
        if not self.nal_queue.full():
            self.nal_queue.put(frame)

    def run(self):
        start_time = time.time()
        frame_count = 0
        fps = self.framerate
        while True:
            frame_count += 1
            elapsed_time = time.time() - start_time
            if elapsed_time > 1 / fps:
                buf = self.nal_queue.get()
                current_fps = frame_count / elapsed_time
                # print(f'Sync Loop FPS: {elapsed_time:.2f}')
                start_time = time.time()
                frame_count = 0
                if buf is not None or buf:
                    self.onframe(self.gsdbs, self.target, buf, self.framerate)


class GSMediaConsumer:

    def __init__(self, gsdbs, target, onframe, framerate):
        self.__tracks = {}
        self.gsdbs = gsdbs
        self.target = target
        self.onframe = onframe
        self.nal_queue = Queue(3)
        self.broadcastthread = FrameBufferThread(self.gsdbs, self.target, self.onframe, framerate)
        self.broadcastthread.start()

    def addTrack(self, track):
        if track not in self.__tracks:
            self.__tracks[track] = None

    async def start(self):
        """
        Start discarding media.
        """
        for track, task in self.__tracks.items():
            if task is None:
                self.__tracks[track] = asyncio.ensure_future(
                    self.stream_consume(track, self.gsdbs, self.target, self.onframe))

    async def stream_consume(self, track, gsdbs, target, onframe):
        while True:
            try:
                frame = await track.recv()
                self.broadcastthread.write(frame)
            except MediaStreamError:
                return

    async def stop(self):
        """
        Stop discarding media.
        """
        for task in self.__tracks.values():
            if task is not None:
                task.cancel()
        self.__tracks = {}


class GSMediaConsumerAsync:

    def __init__(self, gsdbs, target, onframe, frame_queue):
        self.__tracks = {}
        self.gsdbs = gsdbs
        self.target = target
        self.onframe = onframe
        self.frame_queue = frame_queue

    def addTrack(self, track):
        if track not in self.__tracks:
            self.__tracks[track] = None

    async def start(self):
        """
        Start discarding media.
        """
        for track, task in self.__tracks.items():
            if task is None:
                self.__tracks[track] = asyncio.ensure_future(
                    self.stream_consume(track, self.gsdbs, self.target, self.onframe))

    async def stream_consume(self, track, gsdbs, target, onframe):
        while True:
            try:
                frame = await track.recv()
                await self.frame_queue.put(frame)
            except MediaStreamError:
                return

    async def stop(self):
        for task in self.__tracks.values():
            if task is not None:
                task.cancel()
        self.__tracks = {}



import ast
import json
import logging
import platform
from aiortc import VideoStreamTrack
from av import VideoFrame
# import pyaudio as pyaudio
from datetime import datetime
import socketio
from aiortc import RTCPeerConnection, RTCRtpSender, RTCConfiguration, RTCIceServer, RTCSessionDescription
from aiortc.contrib.media import MediaPlayer, MediaRelay, MediaBlackhole

import asyncio
import collections
import time

pcs = set()
relay = None
webcam = None
import numpy as np

_logger = logging.getLogger(__name__)
# p = pyaudio.PyAudio()
# stream = p.open(format=pyaudio.paFloat32,
#                 channels=1,
#                 rate=48000,
#                 output=True)

video_tracklist = []


class FPS:
    def __init__(self, avarageof=50):
        self.frametimestamps = collections.deque(maxlen=avarageof)

    def __call__(self):
        self.frametimestamps.append(time.time())
        if len(self.frametimestamps) > 1:
            return len(self.frametimestamps) / (self.frametimestamps[-1] - self.frametimestamps[0])
        else:
            return 0.0


async def checkTerminatorQueue(opqueue):
    while True:

        data = await opqueue.get()
        if data is not None:
            for track in video_tracklist:
                await track.write(data, 0)


async def checkCommandQueue(commandqueue, channellist):
    while True:
        data = await commandqueue.get()
        if data is not None:
            for channel in channellist:
                if channel.readyState == "open":
                    channel.send(data)
                if channel.readyState != "open" and channel.readyState != "connecting":
                    _logger.warning("closing datachannel : " + str(channel.readyState))
                    channellist.remove(channel)


class TerminatorStreamTrack(VideoStreamTrack):
    """
    A video track that returns an animated flag.
    """

    def __init__(self):
        super().__init__()  # don't forget this!
        self.counter = 0
        height, width = 480, 640
        self.framerate = 0
        self.nal_queue = asyncio.Queue(3)

    async def write(self, frame, framerate=0):
        self.framerate = framerate
        if not self.nal_queue.full():
            await self.nal_queue.put(frame)

    async def recv(self):
        pts, time_base = await self.next_timestamp()

        frame = VideoFrame.from_ndarray(await self.nal_queue.get())
        frame.pts = pts
        frame.time_base = time_base
        self.counter += 1
        return frame


def create_local_tracks(play_from, decode, device, rtbufsize):
    global relay, webcam

    if play_from:
        player = MediaPlayer(play_from, decode=decode)
        return player.audio, player.video
    else:
        options = {"framerate": "30", "video_size": "640x480", "rtbufsize": rtbufsize}
        if relay is None:
            if platform.system() == "Darwin":
                webcam = MediaPlayer(
                    "default:none", format="avfoundation", options=options
                )
            elif platform.system() == "Windows":
                webcam = MediaPlayer(
                    f"video={device}", format="dshow", options=options
                )
            else:
                webcam = MediaPlayer(f"{device}", format="v4l2", options=options)
            relay = MediaRelay()
        return None, relay.subscribe(webcam.video)


def force_codec(pc, sender, forced_codec):
    kind = forced_codec.split("/")[0]
    codecs = RTCRtpSender.getCapabilities(kind).codecs
    transceiver = next(t for t in pc.getTransceivers() if t.sender == sender)
    transceiver.setCodecPreferences(
        [codec for codec in codecs if codec.mimeType == forced_codec]
    )


# class MicStreamTrack(MediaStreamTrack):
#     """
#     An audio stream object for the mic audio from the client
#     """
#     kind = "audio"
#
#     def __init__(self, track):
#         super().__init__()
#         self.track = track
#         self._logger = logging.getLogger(__name__)
#
#     async def recv(self):
#         self._logger.error("mic recv()")
#
#         # Get a new PyAV frame
#         frame = await self.track.recv()
#
#         # Convert to float32 numpy array
#         floatArray = frame.to_ndarray(format="float32")
#         samples = np.sin(np.arange(50000) / 20)
#         # stream.write(floatArray.tostring())
#         stream.write(samples.astype(np.float32).tostring())
#
#         # Put these samples into the mic queue
#         # micSampleQueue.put_nowait(floatArray)
#
#         self._logger.error("Put {} samples to mic queue".format(len(floatArray)))


class GSPeerConnectionBroadcasterUniversal:

    @classmethod
    async def create(cls, gsdbs, target, videostream=True, addChannel=None, oncommand=None):
        self = GSPeerConnectionBroadcasterUniversal()
        self.rtconfiList = []
        self.gsdbs = gsdbs
        self.target = target
        self.videostream = videostream
        self.oncommand = oncommand
        self.addChannel = addChannel

        if self.gsdbs.credentials["stunenable"]:
            self.rtconfiList.append(RTCIceServer(self.gsdbs.credentials["stunserver"]))
        if self.gsdbs.credentials["turnenable"]:
            self.rtconfiList.append(RTCIceServer(urls=self.gsdbs.credentials["turnserver"],
                                                 username=self.gsdbs.credentials["turnuser"],
                                                 credential=self.gsdbs.credentials["turnpw"]
                                                 ))

        self.sio = socketio.AsyncClient()
        self.peerConnections = {}
        self._logger = logging.getLogger(__name__)
        self.webcam = None
        self.relay = None

        @self.sio.event
        async def connect():
            self._logger.info('connection established')

        @self.sio.event
        async def joined(id):
            await self.sio.emit("broadcaster", id)

        @self.sio.event
        async def watcher(id, description):
            if type(description) == str:
                description = ast.literal_eval(description)
            desc = type('new_dict', (object,), description)

            offer = RTCSessionDescription(sdp=description["sdp"], type=description["type"])

            if len(self.rtconfiList) > 0:
                pc = RTCPeerConnection(configuration=RTCConfiguration(self.rtconfiList))
            else:
                pc = RTCPeerConnection()

            pcs.add(pc)

            @pc.on("connectionstatechange")
            async def on_connectionstatechange():
                print("Connection state is %s" % pc.connectionState)
                if pc.connectionState == "failed":
                    await pc.close()
                    pcs.discard(pc)

            @pc.on("track")
            async def on_track(track):
                if track.kind == "video":
                    mediablackhole = MediaBlackhole()
                    mediablackhole.addTrack(track)

            if self.addChannel is not None:
                sendchannel = pc.createDataChannel("terminatorview")
                self.addChannel(sendchannel)

            if videostream:
                audio, video = create_local_tracks(False,
                                                   decode=not True,
                                                   device=self.gsdbs.credentials["webcam"],
                                                   rtbufsize=self.gsdbs.credentials["rtbufsize"])
                if video:
                    video_sender = pc.addTrack(video)
                    force_codec(pc, video_sender, "video/H264")

            await pc.setRemoteDescription(offer)

            answer = await pc.createAnswer()
            await pc.setLocalDescription(answer)

            await self.sio.emit("answer", {"id": id,
                                           "message": json.dumps(
                                               {"type": pc.localDescription.type,
                                                "sdp": pc.localDescription.sdp})})

        @self.sio.event
        async def command(commandcall):
            self.oncommand(self.gsdbs, commandcall)

        @self.sio.event
        async def disconnectPeer(id):
            if id in self.peerConnections:
                await self.peerConnections[id].close()
                self.peerConnections.pop(id, None)

        @self.sio.event
        async def disconnect():
            self._logger.info('disconnected from server')

        connectURL = ""

        if "localhost" in self.gsdbs.credentials["signalserver"]:
            connectURL = f'{self.gsdbs.credentials["signalserver"]}:{str(self.gsdbs.credentials["signalport"])}'
        else:
            connectURL = self.gsdbs.credentials["signalserver"]

        await self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}{self.target}')
        await self.sio.wait()


def channel_log(logger, channel, t, message):
    logger.info("channel(%s) %s %s" % (channel.label, t, message))


def channel_send(logger, channel, message):
    channel_log(logger, channel, ">", message)
    channel.send(message)


class GSPeerConnectionBroadcasterUniversalNew:

    @classmethod
    async def create(cls,
                     gsdbs,
                     target,
                     videostream=True,
                     addChannel=None,
                     oncommand=None,
                     videotrack=None,
                     signalingserver=None
                     ):
        self = GSPeerConnectionBroadcasterUniversalNew()
        self.addChannel = addChannel
        self.rtconfiList = []
        self.target = target
        self.gsdbs = gsdbs
        self.videostream = videostream
        self.oncommand = oncommand
        self.videotrack = videotrack
        self.signalingserver = signalingserver

        if self.gsdbs.credentials["stunenable"]:
            self.rtconfiList.append(RTCIceServer(self.gsdbs.credentials["stunserver"]))
        if self.gsdbs.credentials["turnenable"]:
            self.rtconfiList.append(RTCIceServer(urls=self.gsdbs.credentials["turnserver"],
                                                 username=self.gsdbs.credentials["turnuser"],
                                                 credential=self.gsdbs.credentials["turnpw"]
                                                 ))

        self.sio = socketio.AsyncClient()
        self.peerConnections = {}
        self._logger = logging.getLogger(__name__)
        self.webcam = None
        self.relay = None

        @self.sio.event
        async def connect():
            self._logger.info('connection established')

        @self.sio.event
        async def wjoined(id):
            await self.sio.emit("joined", self.sio.sid)

        @self.sio.event
        async def broadcaster(id, description, emitter):

            if emitter == "watcher":
                return

            if len(self.rtconfiList) > 0:
                pc = RTCPeerConnection(configuration=RTCConfiguration(self.rtconfiList))
            else:
                pc = RTCPeerConnection()
            self.peerConnections[id] = pc

            @pc.on("track")
            async def on_track(track):
                if track.kind == "video":
                    mediablackhole = MediaBlackhole()
                    mediablackhole.addTrack(track)
                # if track.kind == "audio":
                #     micTrack = MicStreamTrack(track)
                #     blackHole = MediaBlackhole()
                #     blackHole.addTrack(micTrack)
                #     await blackHole.start()

            async def on_iceconnectionstatechange():
                print("ICE connection state is ", self.peerConnections[id].iceConnectionState)
                if self.peerConnections[id].iceConnectionState == "failed":
                    await self.peerConnections[id].close()
                if self.peerConnections[id].iceConnectionState == "completed":
                    print("completed")

            self.peerConnections[id]._add_event_handler("iceconnectionstatechange", on_iceconnectionstatechange,
                                                        on_iceconnectionstatechange)

            @pc.on("signalingstatechange")
            async def signalingchange():
                print("change" + pc.signalingState)

            if self.videostream:
                audio, video = create_local_tracks(False,
                                                   decode=not True,
                                                   device=self.gsdbs.credentials["webcam"],
                                                   rtbufsize=self.gsdbs.credentials["rtbufsize"])
                if video:
                    video_sender = pc.addTrack(video)
                    force_codec(pc, video_sender, "video/H264")
            if self.videotrack:
                terminatorviewtrack = TerminatorStreamTrack()
                video_tracklist.append(terminatorviewtrack)
                pc.addTrack(terminatorviewtrack)

            if self.addChannel is not None:
                sendchannel = pc.createDataChannel("terminatorview")
                self.addChannel(sendchannel)
                channel_log(self._logger, sendchannel, "-", "created by local party")

                @sendchannel.on("open")
                def on_open():
                    self._logger.info("datachannel open")

                @sendchannel.on("message")
                def on_message(message):
                    def subprocess_cmd(command):
                        process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
                        proc_stdout = process.communicate()[0].strip()
                        # for line in proc_stdout.decode().split('\n'):
                        channel_send(self._logger, sendchannel, json.dumps({"response": proc_stdout.decode()}))

                    channel_log(self._logger, sendchannel, "<", message)
                    if isinstance(message, str) and message.startswith("ping"):
                        channel_send(self._logger, sendchannel, f'{{"ping":"pong","ts":"{datetime.utcnow()} UTC"}}')
                    if isinstance(message, str) and message.startswith('{"type'):
                        self.oncommand(self.gsdbs, json.loads(message))
                    if isinstance(message, str) and message.startswith('{"command'):
                        subprocess_cmd(json.loads(message)["command"])

            await pc.setLocalDescription(await pc.createOffer())
            await self.sio.emit("watcher", {"target": description,
                                            "id": id,
                                            "sdp": json.dumps(
                                                {"type": pc.localDescription.type,
                                                 "sdp": pc.localDescription.sdp})})

        @self.sio.event
        async def answer(id, description):
            if id in self.peerConnections:
                if self.peerConnections[id].signalingState != "stable":
                    if type(description) == str:
                        description = ast.literal_eval(description)
                    desc = type('new_dict', (object,), description)
                    answer = RTCSessionDescription(sdp=description["sdp"], type=description["type"])
                    await self.peerConnections[id].setRemoteDescription(answer)

        @self.sio.event
        async def disconnectBroadcaster(id, desc):
            if id in self.peerConnections:
                self._logger.info("Disconnected Peer")
                await self.peerConnections[id].close()
                self.peerConnections.pop(id, None)

        @self.sio.event
        async def disconnect():
            self._logger.info('disconnected from server')

        connectURL = ""

        if self.signalingserver is not None:
            connectURL = self.signalingserver

        else:
            if "localhost" in self.gsdbs.credentials["signalserver"]:
                connectURL = f'{self.gsdbs.credentials["signalserver"]}:{str(self.gsdbs.credentials["signalport"])}'
            else:
                connectURL = self.gsdbs.credentials["signalserver"]

        await self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}&target={self.target}&terminator=true')
        await self.sio.wait()


async def createTerminatorView(gsdbs, aioqueue: asyncio.Queue,
                               commandqueue: asyncio.Queue,
                               target,
                               signalingserver=None, oncommand=None):
    datachannelist = []

    def addDatachannel(channel):
        datachannelist.append(channel)

    terminatorviewtask = await asyncio.gather(
        checkCommandQueue(commandqueue, datachannelist),
        checkTerminatorQueue(aioqueue),
        GSPeerConnectionBroadcasterUniversalNew.create(gsdbs=gsdbs,
                                                       target=target + "terminatorview",
                                                       videostream=False,
                                                       videotrack=True,
                                                       signalingserver=signalingserver,
                                                       addChannel=addDatachannel,
                                                       oncommand=oncommand
                                                       ))
    return terminatorviewtask



import logging
import time
import socketio
import ubjson
from multiprocessing.dummy import Pool
import msgpack
import requests

class GSCNodeFunctionReceiverSocket:
    @classmethod
    def create(cls, sio, gsdbs, oncnodefunction, poolsize):
        self = GSCNodeFunctionReceiverSocket()
        self.sio = sio
        self.gsdbs = gsdbs
        self._pool = Pool(self.gsdbs.credentials["poolsize"])
        self.oncnodefunction = oncnodefunction
        self._logger = logging.getLogger(__name__)
        self.poolsize = poolsize

        @self.sio.event
        def connect():
            self._logger.info('oncnodefunction connected')

        def on_success(r):
            pass

        def on_errorPost(error):
            self._logger.exception('cnodefunction failed :' + error)

        @self.sio.event
        def oncnodefunction(id, msg):
            msg1 =  msgpack.unpackb(msg, raw=False)

            self._pool.apply_async(self.oncnodefunction, args=[self.gsdbs, id, msg1],
                                   callback=on_success,
                                   error_callback=on_errorPost)

        connectURL = ""

        if "localhost" in self.gsdbs.credentials["signalserver"]:
            connectURL = f'{self.gsdbs.credentials["signalserver"]}:{str(self.gsdbs.credentials["signalport"])}'
        else:
            connectURL = self.gsdbs.credentials["signalserver"]

        self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}&global=true&poolsize={self.poolsize}')
        self.sio.wait()

    def sendcnodefunctionResult(self, id, msg):
        self.sio.emit("cnodeanswer", id, msg)


class GSCNodeFunctionReceiver:
    def __init__(self, gsdbs, oncnodefunction, poolsize):
        self.sio = socketio.Client()
        self.gsdbs = gsdbs
        self.oncnodefunction = oncnodefunction
        self.poolsize = poolsize

    def startSocket(self):
        GSCNodeFunctionReceiverSocket.create(self.sio, self.gsdbs, self.oncnodefunction, self.poolsize)

    def sendcnodefunctionanswer(self, id, msg):

        encoded = msgpack.packb(msg, use_bin_type=True)

        self.sio.emit("cnodeanswer", {"id": id, "data": encoded})


class GSCNodeFunctionCallerSocket:
    @classmethod
    def create(cls, sio, gsdbs, oncnodefunctionanswer,loadbalancerurl):
        self = GSCNodeFunctionCallerSocket()
        self.sio = sio
        self.gsdbs = gsdbs
        self.loadbalancerurl=loadbalancerurl
        self._logger = logging.getLogger(__name__)
        self.oncnodefunctionanswer = oncnodefunctionanswer

        @self.sio.event
        def connect():
            self._logger.warning('oncnodefunction connected')

        @self.sio.event
        def cnodeanswer(id, message):
            if (type(message) is dict):
                self.oncnodefunctionanswer(self.gsdbs, message)
            else:
                self.oncnodefunctionanswer(self.gsdbs, msgpack.unpackb(message, raw=False))

        connectURL = ""

        if "localhost" in self.loadbalancerurl:
            connectURL = f'{self.loadbalancerurl}'
        else:
            connectURL = self.loadbalancerurl

        self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}&caller=true')
        self.sio.wait()


class GSCNodeFunctionCaller:
    def __init__(self, gsdbs, oncnodefunctionanswer,loadbalancerurl,poolsize):
        self.sio = socketio.Client()
        self.gsdbs = gsdbs
        self.oncnodefunctionanswer = oncnodefunctionanswer
        self._pool = Pool(poolsize)
        self.loadbalancerurl=loadbalancerurl
        self._logger = logging.getLogger(__name__)

    def startSocket(self):
        GSCNodeFunctionCallerSocket.create(self.sio,
                                           self.gsdbs,
                                           self.oncnodefunctionanswer,
                                           self.loadbalancerurl)

    def on_success(self, r):
        self._logger.debug('cnodefunction succeed')

    def on_errorPost(self, error):
        self._logger.exception('cnodefunction failed :' + error)

    def emitFunction(self, sio, target, msg):
        msg["cnodefunction"] = target
        encoded = msgpack.packb(msg, use_bin_type=True)
        sio.emit(target, encoded)

    def sendcnodefunction(self, target, msg):
        time.sleep(self.gsdbs.credentials["cnodesendsleep"])
        self._pool.apply_async(self.emitFunction,
                               args=[self.sio, target, msg],
                               callback=self.on_success,
                               error_callback=self.on_errorPost)
class GSCNodeFunctionCallerRest:
    def __init__(self, gsdbs, oncnodefunctionanswer):

        self.gsdbs = gsdbs
        self.oncnodefunctionanswer = oncnodefunctionanswer
        self._pool = Pool(self.gsdbs.credentials["poolsize"])
        self._logger = logging.getLogger(__name__)

    def on_success(self, r):
        self._logger.debug('cnodefunction succeed')

    def on_errorPost(self, error):
        self._logger.exception('cnodefunction failed :' + str(error))

    def emitFunction(self, target, msg, timeout):
        msg["cnodefunction"] = target

        encoded = ubjson.dumpb(msg)
        try:
            r = requests.post(target, data=encoded, timeout=timeout)
            self.oncnodefunctionanswer(self.gsdbs, ubjson.loadb(r.content))
        except requests.exceptions.Timeout:
            self._logger.exception("timeout")
            self.oncnodefunctionanswer(self.gsdbs, {"error": "true",
                                                    "message": "capacity reached"})
        except requests.exceptions.RequestException as e:
            self._logger.exception("something went very wrong" + str(e))
            self.oncnodefunctionanswer(self.gsdbs, {"error": "true",
                                                    "message": "capacity reached"})

        # sio.emit("oncnodefunction", {"target": target, "data": encoded})

    def sendcnodefunction(self, target, msg, timeout):
        self._pool.apply_async(self.emitFunction,
                               args=[target, msg, timeout],
                               callback=self.on_success,
                               error_callback=self.on_errorPost)


import ast
import json
import logging
from threading import Thread
from time import sleep
import socketio
from aiortc import RTCPeerConnection, RTCRtpSender, RTCConfiguration, RTCIceServer, RTCSessionDescription
from aiortc.contrib.media import MediaPlayer, MediaRelay

try:
    import picamera
except:
    log = logging.getLogger(__name__)
    log.error("failed to import picamera")

from gspeerconnection.pitrack import PiH264Relay, PiH264CameraOutput

video_trackList = list()
pcs = set()
relay = None
microfon = None


class BroadcastThread(Thread):
    def __init__(self, cameraoutput):
        super(BroadcastThread, self).__init__()
        self.cameraoutput = cameraoutput

    def run(self):
        while True:
            buf = self.cameraoutput.nal_queue.get()
            if buf:
                for track in video_trackList:
                    track.write(buf)


class GSPeerConnectionBroadcasterPi:

    def getPiH264Relay(self):
        video_track = PiH264Relay(30)
        video_trackList.append(video_track)
        return video_track

    def getMicrofonRelay(self):
        global relay, microfon
        if relay is None:
            microfon = MediaPlayer("hw:1", format="alsa", options={
                "channels": "1",
                "sample_rate": "44100"
            })
            relay = MediaRelay()
        return relay.subscribe(microfon.audio)

    @classmethod
    async def create(cls, gsdbs, camera, oncommand=None):
        self = GSPeerConnectionBroadcasterPi()
        self.rtconfiList = []

        self.gsdbs = gsdbs
        self.oncommand = oncommand
        # picam.append(camera)
        if self.gsdbs.credentials["stunenable"]:
            self.rtconfiList.append(RTCIceServer(self.gsdbs.credentials["stunserver"]))
        if self.gsdbs.credentials["turnenable"]:
            for turnserver in self.gsdbs.credentials["turnserver"]:
                self.rtconfiList.append(RTCIceServer(urls=turnserver,
                                                     username=self.gsdbs.credentials["turnuser"],
                                                     credential=self.gsdbs.credentials["turnpw"]
                                                     ))

        camera.resolution = (self.gsdbs.credentials["hres"], self.gsdbs.credentials["vres"])
        camera.framerate = self.gsdbs.credentials["framerate"]
        sleep(1)  # camera warm-up time
        target_bitrate = camera.resolution[0] * \
                         camera.resolution[1] * \
                         camera.framerate * 0.150

        cameraOutput = PiH264CameraOutput()
        broadcast_thread = BroadcastThread(cameraOutput)
        camera.start_recording(
            cameraOutput,
            format="h264",
            profile="constrained",
            bitrate=int(target_bitrate),
            inline_headers=True,
            sei=False,
        )

        broadcast_thread.start()
        self.sio = socketio.AsyncClient()
        self.peerConnections = {}
        self._logger = logging.getLogger(__name__)
        self.webcam = None
        self.relay = None

        @self.sio.event
        async def connect():
            self._logger.info('connection established')

        @self.sio.event
        async def watcher(id, description):
            if type(description) == str:
                description = ast.literal_eval(description)
            desc = type('new_dict', (object,), description)

            offer = RTCSessionDescription(sdp=description["sdp"], type=description["type"])

            self._logger.debug(str(self.rtconfiList))

            if len(self.rtconfiList) > 0:
                pc = RTCPeerConnection(configuration=RTCConfiguration(self.rtconfiList))
            else:
                pc = RTCPeerConnection()

            self.peerConnections[id] = pc

            video = self.getPiH264Relay()
            audio = None

            @pc.on("connectionstatechange")
            async def on_connectionstatechange():
                print("Connection state is %s" % pc.connectionState)
                if pc.connectionState == "failed":
                    await pc.close()
                    self.peerConnections.pop(id, None)

            transceiver = pc.addTransceiver("video", direction="sendonly")

            capabilities = RTCRtpSender.getCapabilities("video")
            preferences = list(filter(lambda x: x.name == "H264", capabilities.codecs))
            preferences += list(filter(lambda x: x.name == "rtx", capabilities.codecs))
            transceiver.setCodecPreferences(preferences)

            if self.gsdbs.credentials["audio"]:
                pc.addTransceiver("audio", direction="sendonly")
                audio = self.getMicrofonRelay()
                pc.addTrack(audio)

            for t in pc.getTransceivers():
                if t.kind == "video":
                    pc.addTrack(video)

            await pc.setRemoteDescription(offer)

            answer = await pc.createAnswer()
            await pc.setLocalDescription(answer)

            await self.sio.emit("answer", {"id": id,
                                           "message": json.dumps(
                                               {"type": pc.localDescription.type,
                                                "sdp": pc.localDescription.sdp})})

        @self.sio.event
        async def command(commandcall):
            self.oncommand(self.gsdbs, commandcall)

        @self.sio.event
        async def joined(id):
            await self.sio.emit("broadcaster", id)

        @self.sio.event
        async def disconnectPeer(id):
            if id in self.peerConnections:
                await self.peerConnections[id].close()
                self.peerConnections.pop(id, None)

        @self.sio.event
        async def candidate(id, candidate):
            await self.peerConnections[id].addIceCandidate()

        @self.sio.event
        async def disconnect():
            self._logger.info('disconnected from server')

        connectURL = ""

        if "localhost" in self.gsdbs.credentials["signalserver"]:
            connectURL = f'{self.gsdbs.credentials["signalserver"]}:{str(self.gsdbs.credentials["signalport"])}'
        else:
            connectURL = self.gsdbs.credentials["signalserver"]

        await self.sio.connect(
            f'{connectURL}?gssession={self.gsdbs.cookiejar.get("session")}.{self.gsdbs.cookiejar.get("signature")}{self.gsdbs.credentials["cnode"]}')
        await self.sio.wait()



import asyncio
import logging
import math
from queue import Queue
from struct import pack
from subprocess import Popen
from threading import Thread
from typing import Iterator, List, Tuple
try:
    from aiortc.mediastreams import EncodedMediaStreamTrack
except:
    log = logging.getLogger(__name__)
    log.error("wrong aiortc version")

PACKET_MAX = 1300

NAL_TYPE_FU_A = 28
NAL_TYPE_STAP_A = 24

NAL_HEADER_SIZE = 1
FU_A_HEADER_SIZE = 2
LENGTH_FIELD_SIZE = 2
STAP_A_HEADER_SIZE = NAL_HEADER_SIZE + LENGTH_FIELD_SIZE
logger = logging.getLogger(__name__)

class PiH264CameraOutput(object):
    def __init__(self):
        logger.debug('Spawning background buffer process')
        self.nal_queue = Queue(3)
        self.nal_buffer = None

    def write(self, buf):
        if self.nal_buffer is None:
            self.nal_buffer = buf
        else:
            self.nal_buffer += buf
        if (
                len(self.nal_buffer) < 64
        ):  # Just making sure to pack SPS/PPS within a single buffer
            return
        if not self.nal_queue.full():
            self.nal_queue.put(self.nal_buffer)
        self.nal_buffer = None

    def flush(self):
        logger.debug('Waiting for background conversion process to exit')




class PiH264Relay(EncodedMediaStreamTrack):

    kind = "video"
    _start: float
    _timestamp: int

    def __init__(self, video_rate, clock_rate=90000) -> None:
        super().__init__()
        self.nal_queue = Queue(3)
        self._timestamp = 0
        self._frame_time = 1 / video_rate
        self._clock_rate = clock_rate
        self.nal_buffer = None

    def write(self, buf: bytes):
        if self.nal_buffer is None:
            self.nal_buffer = buf
        else:
            self.nal_buffer += buf
        if (
            len(self.nal_buffer) < 64
        ):  # Just making sure to pack SPS/PPS within a single buffer
            return
        if not self.nal_queue.full():
            self.nal_queue.put(self.nal_buffer)
        self.nal_buffer = None

    @staticmethod
    def _packetize_fu_a(data: bytes) -> List[bytes]:
        available_size = PACKET_MAX - FU_A_HEADER_SIZE
        payload_size = len(data) - NAL_HEADER_SIZE
        num_packets = math.ceil(payload_size / available_size)
        num_larger_packets = payload_size % num_packets
        package_size = payload_size // num_packets

        f_nri = data[0] & (0x80 | 0x60)  # fni of original header
        nal = data[0] & 0x1F

        fu_indicator = f_nri | NAL_TYPE_FU_A

        fu_header_end = bytes([fu_indicator, nal | 0x40])
        fu_header_middle = bytes([fu_indicator, nal])
        fu_header_start = bytes([fu_indicator, nal | 0x80])
        fu_header = fu_header_start

        packages = []
        offset = NAL_HEADER_SIZE
        while offset < len(data):
            if num_larger_packets > 0:
                num_larger_packets -= 1
                payload = data[offset : offset + package_size + 1]
                offset += package_size + 1
            else:
                payload = data[offset : offset + package_size]
                offset += package_size

            if offset == len(data):
                fu_header = fu_header_end

            packages.append(fu_header + payload)

            fu_header = fu_header_middle
        assert offset == len(data), "incorrect fragment data"

        return packages

    @staticmethod
    def _packetize_stap_a(
        data: bytes, packages_iterator: Iterator[bytes]
    ) -> Tuple[bytes, bytes]:
        counter = 0
        available_size = PACKET_MAX - STAP_A_HEADER_SIZE

        stap_header = NAL_TYPE_STAP_A | (data[0] & 0xE0)

        payload = bytes()
        try:
            nalu = data  # with header
            while len(nalu) <= available_size and counter < 9:
                stap_header |= nalu[0] & 0x80

                nri = nalu[0] & 0x60
                if stap_header & 0x60 < nri:
                    stap_header = stap_header & 0x9F | nri

                available_size -= LENGTH_FIELD_SIZE + len(nalu)
                counter += 1
                payload += pack("!H", len(nalu)) + nalu
                nalu = next(packages_iterator)

            if counter == 0:
                nalu = next(packages_iterator)
        except StopIteration:
            nalu = None

        if counter <= 1:
            return data, nalu
        else:
            return bytes([stap_header]) + payload, nalu

    @classmethod
    def _packetize(cls, packages: Iterator[bytes]) -> List[bytes]:
        packetized_packages = []

        packages_iterator = iter(packages)
        package = next(packages_iterator, None)
        while package is not None:
            if len(package) > PACKET_MAX:
                packetized_packages.extend(cls._packetize_fu_a(package))
                package = next(packages_iterator, None)
            else:
                packetized, package = cls._packetize_stap_a(package, packages_iterator)
                packetized_packages.append(packetized)
        return packetized_packages

    @staticmethod
    def _split_bitstream(buf: bytes) -> Iterator[bytes]:
        # TODO: write in a more pytonic way,
        # translate from: https://github.com/aizvorski/h264bitstream/blob/master/h264_nal.c#L134
        i = 0
        while True:
            while (buf[i] != 0 or buf[i + 1] != 0 or buf[i + 2] != 0x01) and (
                buf[i] != 0 or buf[i + 1] != 0 or buf[i + 2] != 0 or buf[i + 3] != 0x01
            ):
                i += 1  # skip leading zero
                if i + 4 >= len(buf):
                    return
            if buf[i] != 0 or buf[i + 1] != 0 or buf[i + 2] != 0x01:
                i += 1
            i += 3
            nal_start = i
            if (buf[nal_start] & 0x1F) != 7 and (
                buf[nal_start] & 0x1F
            ) != 8:  # Assuming only SPS/PPS are packaged with other NALs
                yield buf[nal_start:]
                return
            while (buf[i] != 0 or buf[i + 1] != 0 or buf[i + 2] != 0) and (
                buf[i] != 0 or buf[i + 1] != 0 or buf[i + 2] != 0x01
            ):
                i += 1
                # FIXME: the next line fails when reading a nal that ends
                # exactly at the end of the data
                if i + 3 >= len(buf):
                    nal_end = len(buf)
                    yield buf[nal_start:]
                    return  # did not find nal end, stream ended first
            nal_end = i
            yield buf[nal_start:nal_end]

    async def recv_encoded(self, keyframe=False) -> List[bytes]:
        while True:
            if self.nal_queue.empty():
                await asyncio.sleep(self._frame_time)
                continue
            nal = self.nal_queue.get()
            #if (nal[4] & 0x1F) != 0x01 or not keyframe:
            #    break
            break
        self._timestamp += int(self._frame_time * self._clock_rate)
        timestamp = self._timestamp
        return self._packetize(self._split_bitstream(nal)), timestamp

